/* --- Copyright University of Sussex 1999. All rights reserved. ----------
 * File:			C.win32/extern/unicode/encoding_windows.c
 * Purpose:			Unicode <-> Windows ANSI Code Page
 * Author:			Robert Duncan, Oct 28 1999 (see revisions)
 */

#include <windows.h>
#include "encoding.h"

static UNICODE_TO_X unicode_to_ansi;
static X_TO_UNICODE ansi_to_unicode;

CODING_PAIR encoding_windows = { unicode_to_ansi, ansi_to_unicode };

static BOOL init_done = FALSE;
static WORD cpid = CP_ACP;
static BOOL double_byte = FALSE;
static BYTE default_char[MAX_DEFAULTCHAR] = { '?' };
static DWORD default_char_len = 1;
static BOOL ascii = TRUE;

__declspec(dllexport)
int encoding_windows_init(unsigned codepage) {
	/*	Cache information about the current code page.
		NB: this is strictly incorrect because the code page could change
		dynamically, but it's too expensive to recompute every time.
		We also assume that this is only ever called from the same thread,
		which is true for Poplog.
	*/
	LCID locale;
	CPINFO cpinfo;
	DWORD nbytes;
	char info[16];
	BYTE c;
	if (codepage == CP_ACP) {
		/* Get the ANSI codepage identifier for the calling thread */
		locale = GetThreadLocale();
		nbytes = GetLocaleInfo(locale, LOCALE_IDEFAULTANSICODEPAGE, info,
				sizeof info);
		if (nbytes > 0 && nbytes < sizeof info) {
			codepage = (unsigned)strtoul(info, 0, 10);
		}
	}
	/* Get character set information from the code page */
	if (!GetCPInfo(codepage, &cpinfo)) {
		return 0;
	}
	cpid = (WORD)codepage;
	/* check for double-byte character set */
	if (cpinfo.MaxCharSize == 2 &&
			(cpinfo.LeadByte[0] || cpinfo.LeadByte[1])) {
		double_byte = TRUE;
	}
	/* copy the default character */
	strcpy(default_char, cpinfo.DefaultChar);
	default_char_len = strlen(default_char);
	/* Check assumption that the first 128 characters are ASCII */
	for (c = 0; c <= 0x7f; ++c) {
		WCHAR wc;
		if (MultiByteToWideChar(cpid, MB_ERR_INVALID_CHARS, &c, 1, &wc, 1) == 0
				|| wc != c) {
			ascii = FALSE;
			break;
		}
	}
	init_done = TRUE;
	return 1;
}


/***
 *** NB: These functions are heavily optimised for typical usage cases
 ***/

static void unicode_to_ansi(
	const CHAR16*	input,
	unsigned*		inputLength,
	CHAR8*			output,
	unsigned*		outputLength,
	unsigned*		state,
	unsigned		flags
) {
	unsigned ilen = *inputLength;
	unsigned olen = *outputLength;

	if (!init_done) encoding_windows_init(CP_ACP);

	if (flags & CHAR8_INPUT) {
		/* Input is 8-bit */
		const CHAR8* charInput = (CHAR8*)input;
		/* Copy a leading sequence of ASCII characters */
		if (ascii) {
			unsigned i, icnt = ilen;
			if (olen < icnt) icnt = olen;
			for (i = 0; i < icnt && charInput[i] <= 0x7f; ++i) {
				output[i] = charInput[i];
				if (output[i] == '\n' && (flags & STOP_AT_NEWLINE)) {
					ilen -= i+1, olen -= i+1;
					goto finish;
				}
			}
			charInput += i, ilen -= i;
			output += i, olen -= i;
		}
		/* Try a block of characters at a time */
		while (ilen > 0 && olen > 0) {
			enum { BLOCK_SIZE = 64 };
			CHAR16 block[BLOCK_SIZE];
			int nbytes;
			unsigned i, icnt = (double_byte ? olen / 2 : olen);
			if (icnt == 0) break;
			if (icnt > BLOCK_SIZE) icnt = BLOCK_SIZE;
			if (icnt > ilen) icnt = ilen;
			for (i = 0; i < icnt; ++i) {
				block[i] = charInput[i];
				if (block[i] == L'\n' && (flags & STOP_AT_NEWLINE)) {
					icnt = i + 1;
					break;
				}
			}
			nbytes = WideCharToMultiByte(cpid, 0, block, icnt, output, olen,
					NULL, NULL);
			if (nbytes == 0) {
				break;
			}
			else if (block[icnt-1] == L'\n' && (flags & STOP_AT_NEWLINE)) {
				ilen -= icnt, olen -= nbytes;
				goto finish;
			}
			else {
				charInput += icnt, ilen -= icnt;
				output += nbytes, olen -= nbytes;
			}
		}
		/* Do remainder one at a time */
		while (ilen > 0 && olen > 0) {
			CHAR16 c = charInput[0];
			if (ascii && c <= 0x7f) {
				output[0] = (CHAR8)c;
				++charInput, --ilen;
				++output, --ilen;
			}
			else {
				int nbytes = WideCharToMultiByte(cpid, 0, &c, 1, output, olen,
						NULL, NULL);
				if (nbytes > 0) {
					++charInput, --ilen;
					output += nbytes, olen -= nbytes;
				}
				else if (GetLastError() == ERROR_INSUFFICIENT_BUFFER ||
						default_char_len > olen) {
					goto finish;
				}
				else {
					strcpy(output, default_char);
					++charInput, --ilen;
					output += default_char_len, olen -= default_char_len;
				}
			}
			if (c == L'\n' && (flags & STOP_AT_NEWLINE)) {
				break;
			}
		}
	}
	else {
		/* Copy a leading sequence of ASCII characters */
		if (ascii) {
			unsigned i, icnt = ilen;
			if (olen < icnt) icnt = olen;
			for (i = 0; i < icnt && input[i] <= 0x7f; ++i) {
				output[i] = (CHAR8)input[i];
				if (output[i] == '\n' && (flags & STOP_AT_NEWLINE)) {
					ilen -= i+1, olen -= i+1;
					goto finish;
				}
			}
			input += i, ilen -= i;
			output += i, olen -= i;
		}
		/* Try a block of characters at a time */
		while (ilen > 0 && olen > 0) {
			int nbytes;
			unsigned i, icnt = (double_byte ? olen / 2 : olen);
			if (icnt > ilen) {
				icnt = ilen;
			}
			else if (icnt == 0) {
				icnt = 1;
			}
			if (flags & STOP_AT_NEWLINE) {
				for (i = 0; i < icnt && input[i] != L'\n'; ++i) {
				}
				if (i < icnt) {
					icnt = i + 1;
				}
			}
			nbytes = WideCharToMultiByte(cpid, 0, input, icnt, output, olen,
					NULL, NULL);
			if (nbytes == 0) {
				break;
			}
			else if (input[icnt-1] == L'\n' && (flags & STOP_AT_NEWLINE)) {
				ilen -= icnt, olen -= nbytes;
				goto finish;
			}
			else {
				input += icnt, ilen -= icnt;
				output += nbytes, olen -= nbytes;
			}
		}
		/* Do remainder one at a time */
		while (ilen > 0 && olen > 0) {
			if (ascii && input[0] <= 0x7f) {
				output[0] = (CHAR8)input[0];
				++output, --ilen;
				++input, --ilen;
			}
			else {
				int nbytes = WideCharToMultiByte(cpid, 0, input, 1, output,
						olen, NULL, NULL);
				if (nbytes > 0) {
					++input, --ilen;
					output += nbytes, olen -= nbytes;
				}
				else if (GetLastError() == ERROR_INSUFFICIENT_BUFFER ||
						default_char_len > olen) {
					goto finish;
				}
				else {
					strcpy(output, default_char);
					++input, --ilen;
					output += default_char_len, olen -= default_char_len;
				}
			}
			if (input[-1] == L'\n' && (flags & STOP_AT_NEWLINE)) {
				break;
			}
		}
	}

finish:
	*inputLength = ilen;
	*outputLength -= olen;
}

static void ansi_to_unicode(
	const CHAR8*	input,
	unsigned*		inputLength,
	CHAR16*			output,
	unsigned*		outputLength,
	unsigned*		state,
	unsigned		flags)
{
	unsigned ilen = *inputLength;
	unsigned olen = *outputLength;

	if (!init_done) encoding_windows_init(CP_ACP);

	if (ascii) {
		/* Copy a leading sequence of ASCII characters */
		unsigned i, icnt = ilen;
		if (olen < icnt) icnt = olen;
		for (i = 0; i < icnt && input[i] <= 0x7f; ++i) {
			output[i] = input[i];
			if (output[i] == L'\n' && (flags & STOP_AT_NEWLINE)) {
				ilen -= i+1, olen -= i+1;
				goto finish;
			}
		}
		input += i, ilen -= i;
		output += i, olen -= i;
	}

	if (ilen == 0 || olen == 0) {
		goto finish;
	}

	if (double_byte) {
		/* Try to do everything in one call */
		DWORD mbFlags = 0;
		if (ascii || !(flags & STOP_AT_NEWLINE)) {
			int nchars;
			unsigned icnt = 0, ocnt = 0;
			while (icnt < ilen && ocnt < olen) {
				if (ascii && input[icnt] <= 0x7f) {
					/* ASCII */
					++icnt, ++ocnt;
					if ((flags & STOP_AT_NEWLINE) && input[icnt-1] == '\n') {
						break;
					}
				}
				else if (!IsDBCSLeadByteEx(cpid, input[icnt])) {
					/* single byte, non-ASCII */
					++icnt, ++ocnt;
				}
				else if (icnt + 1 < ilen) {
					/* double byte */
					icnt += 2, ++ocnt;
					/* check for invalid trail byte */
					mbFlags = MB_ERR_INVALID_CHARS;
				}
				else if (flags & INPUT_INCOMPLETE) {
					/* lead byte to be left unread */
					break;
				}
				else {
					/* lead byte to be treated as error */
					icnt = ocnt = 0;
					break;
				}
			}
			nchars = MultiByteToWideChar(cpid, mbFlags, input, icnt, output,
					olen);
			if (nchars > 0) {
				ilen -= icnt, olen -= nchars;
				goto finish;
			}
		}
		/* Fall back on one character at a time */
		while (ilen > 0 && olen > 0) {
			if (ascii && input[0] <= 0x7f) {
				/* ASCII */
				output[0] = input[0];
				++input, --ilen;
				++output, --olen;
			}
			else if (!IsDBCSLeadByteEx(cpid, input[0])) {
				/* single byte, non-ASCII */
				int nchars = MultiByteToWideChar(cpid, 0, input, 1, output,
						olen);
				if (nchars > 0) {
					++input, --ilen;
					output += nchars, olen -= nchars;
				}
				else if (GetLastError() == ERROR_INSUFFICIENT_BUFFER) {
					goto finish;
				}
				else {
					output[0] = U_REPLACEMENT_CHARACTER;
					++input, --ilen;
					++output, --olen;
				}
			}
			else if (ilen > 1) {
				/* double byte */
				int nchars = MultiByteToWideChar(cpid, MB_ERR_INVALID_CHARS,
						input, 2, output, olen);
				if (nchars > 0) {
					input += 2, ilen -= 2;
					output += nchars, olen -= nchars;
				}
				else if (GetLastError() == ERROR_INSUFFICIENT_BUFFER) {
					goto finish;
				}
				else {
					output[0] = U_REPLACEMENT_CHARACTER;
					input += 2, ilen -= 2;
					++output, --olen;
				}
			}
			else if (flags & INPUT_INCOMPLETE) {
				/* lead byte to be left unread */
				goto finish;
			}
			else {
				/* lead byte to be treated as error */
				output[0] = U_REPLACEMENT_CHARACTER;
				--ilen, --olen;
				goto finish;
			}
			if ((flags & STOP_AT_NEWLINE) && output[-1] == L'\n') {
				break;
			}
		}
	}
	else {
		/* Try to do everything in one call */
		if (!(flags & STOP_AT_NEWLINE)) {
			int nchars;
			unsigned icnt = ilen;
			if (olen < icnt) icnt = olen;
			nchars = MultiByteToWideChar(cpid, 0, input, icnt, output, olen);
			if (nchars > 0) {
				ilen -= icnt, olen -= nchars;
				goto finish;
			}
		}
		else if (ascii) {
			int nchars;
			unsigned i, icnt = ilen;
			if (olen < icnt) icnt = olen;
			for (i = 0; i < icnt && input[i] != '\n'; ++i) {
			}
			if (i < icnt) {
				icnt = i + 1;
			}
			nchars = MultiByteToWideChar(cpid, 0, input, icnt, output, olen);
			if (nchars > 0) {
				ilen -= icnt, olen -= nchars;
				goto finish;
			}
		}
		/* Fall back on one character at a time */
		while (ilen > 0 && olen > 0) {
			int nchars = MultiByteToWideChar(cpid, 0, input, 1, output, olen);
			if (nchars > 0) {
				++input, --ilen;
				output += nchars, olen -= nchars;
			}
			else if (GetLastError() == ERROR_INSUFFICIENT_BUFFER) {
				goto finish;
			}
			else {
				output[0] = U_REPLACEMENT_CHARACTER;
				++input, --ilen;
				++output, --olen;
			}
			if ((flags & STOP_AT_NEWLINE) && output[-1] == L'\n') {
				break;
			}
		}
	}

finish:
	*inputLength = ilen;
	*outputLength -= olen;
}


/* --- Revision History ---------------------------------------------------
--- Robert Duncan, Dec  7 1999
		Exported the init function so that the caller can set the codepage
		instead of always assuming the thread default.
 */
